/**********************************************************************
$Id: highwind.c,v 1.9 2002/10/14 18:24:27 bu Exp $

Copyright 2001-2002, Bulent Yilmaz

Part of Microbrew MicroSieve
USENET SPAM filter

http://www.microbrew.org/products/usieve.html


MicroSieve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MicroSieve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MicroSieve; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

**********************************************************************/

#include <errno.h>
#include "highwind.h"

int endprog=0;

#if LOG_STATS
static long int alarminterval=0;
#endif

void sighandler(int signum) {

    /* Set up the signal handler to do the right thing when it receives
       either one of the following signals:
    
        1 - HUP 
        14 - ALRM (Only for stats logging)
        15 - TERM
        
    */

    FILE *temphandle;
    
    temphandle=getloghandle();
    
    fprintf(temphandle, "Caught signal %d.\n", signum);
    fflush(temphandle);
    
#if LOG_STATS
    statsoutput();
    
    if(!alarminterval) alarminterval=getstatsinterval();

    if(signum==14) alarm(alarminterval);
#endif    

    if(signum==1) endprog=1;
    if(signum==15) endprog=2;

}

int main() {

    char *inbuf=0;
    long int read_index=0, newread=0, max_art_size=0, realartsize=0;
    int reject=0, eom=0, busywait=0;

#if TESTOUTPUT
    FILE *temphandle;
#endif

#if STATLOGGING
    long int articles=0, begintime=0, nowtime=0;
    unsigned long int bytes=0;
#endif

#if LOG_STATS
    struct sigaction sigalrm;
#endif

    struct sigaction sighup, sigterm;

#if LOG_STATS    
    /* Set up the alarm handler. */
    
    memset(&sigalrm, 0, sizeof(sigalrm));
    sigalrm.sa_handler = sighandler;
    
    if( sigaction(SIGALRM, &sigalrm, NULL) ) perror("sigaction");
    alarm(1);
#endif
    
    /* Set up the HUP handler. */
    
    memset(&sighup, 0, sizeof(sighup));
    sighup.sa_handler = sighandler;
    
    if( sigaction(SIGHUP, &sighup, NULL) ) perror("sigaction");
    
    /* Set up the TERM handler. */
    
    memset(&sigterm, 0, sizeof(sigterm));
    sigterm.sa_handler = sighandler;
    
    if( sigaction(SIGTERM, &sigterm, NULL) ) perror("sigaction");


    /* Start up the program */

    cflog("Microbrew MicroSieve " USIEVE_VER " started");

    writepid();
    
    loadtable();
    
    max_art_size=getmaxartsize();
    
    inbuf=malloc(max_art_size);
    if(!inbuf) {
       cflog("Could not allocate buffer for articles.  Ending program.");
       return -1;
    }
    
    /* Magic: */

    setvbuf(stdin, NULL, _IONBF, 0);

    while(!endprog) {
      reject=0;
      read_index=0;
      eom=0;
      busywait=0;

      *inbuf=0;
      while((!eom)&&(!endprog)) {
#if TESTOUTPUT
          cflog("Trying to do a read");
#endif
#if TESTOUTPUT
          temphandle=getloghandle();
          fprintf(temphandle, "Already %ld bytes in buffer.\n",read_index);
          fflush(temphandle);
#endif

          if(busywait) if(getppid()==1) endprog=2;

	  if(feof(stdin)) {
	      clearerr(stdin);
#if TESTOUTPUT
              cflog("Cleared error on stdin");
#endif          	      
	  }
#if TESTOUTPUT
          temphandle=getloghandle();
          fprintf(temphandle, "Trying to read %ld bytes.\n",(max_art_size-read_index));
          fflush(temphandle);
#endif                              

	  newread=read(STDIN_FILENO, (inbuf + read_index), (max_art_size-read_index));

#if TESTOUTPUT
          temphandle=getloghandle();
          fprintf(temphandle, "Just read %ld bytes.\n",newread);
          if (newread < 0) {
              fprintf(temphandle, "read returned errorno %d\n", errno);
          }
          fflush(temphandle);
#endif

          if(newread > 0) read_index+=newread;
          *(inbuf+read_index)='\0';

          if((*(inbuf+read_index-1)=='\n')&&(*(inbuf+read_index-2)=='\r')&&(*(inbuf+read_index-3)=='.')&&(*(inbuf+read_index-4)=='\n')) {
	      eom=1;
	      realartsize=read_index;
#if TESTOUTPUT
	      temphandle=getloghandle();
	      fprintf(temphandle, "Article: %ld bytes.\n",read_index);
	      fflush(temphandle);
#endif
          } else if( read_index >= max_art_size ) {	
          
              // If the message is greater than MAX_ART_SIZE:
              
              realartsize=read_index;
              
              reject=1;
              eom=1;
              
              // Read to end of article and go on.
              
              while( read_index > 0 ) {
 
                  if( !((*(inbuf+read_index-1)=='\n')&&(*(inbuf+read_index-2)=='\r')&&(*(inbuf+read_index-3)=='.')&&(*(inbuf+read_index-4)=='\n')) ) {
#if TESTOUTPUT
		      cflog("Clearing out end of input buffer");
#endif
                      read_index=read(STDIN_FILENO, inbuf, max_art_size);
                      realartsize+=read_index;

		  } else {
#if TESTOUTPUT
                      cflog("Got the end of the article.");
#endif
                      read_index=-2;
                  }
              }
#if TESTOUTPUT
	      cflog("Message too big.  Rejecting...");
#endif
          }
      
          if( (read_index==0) && (!eom) ) busywait=1;
          else busywait=0;
         
      }
      
      // Do actual spam filtering.
      
      if((read_index > 5)&&(!reject)) reject=usieve(inbuf, read_index);
        
      if(!busywait) {
          if(reject>0) {
              fputs(REJECTCODE, stdout);
#if TESTOUTPUT
              cflog("Article Rejected");
#endif
          } else {
              fputs(ACCEPTCODE, stdout);
#if TESTOUTPUT
              cflog("Article Passed.");
#endif
          }
          fflush(stdout);

#if LOG_STATS
          statlog(realartsize, reject);
#endif
      } 
#if TESTOUTPUT      
      else cflog("Caught HUP/TERM in busy-wait state.");      
#endif    

      if(endprog) {
          cflog("Reloading configuration...");

          usieve_reload();

#if LOG_STATS
          alarminterval=0;
          alarm(1);
#endif

          free(inbuf);
        
          if(endprog==1) {
        
              max_art_size=getmaxartsize();
    
              inbuf=malloc(max_art_size);
              if(!inbuf) {
                  cflog("Could not reallocate buffer for articles.  Ending program.");
                  return -1;
              }
              
              endprog=0;
          }
      }
    
    } 
    
    usieve_term();

    cflog("Exited program normally.");
    return 0;

}
