/**********************************************************************
$Id: sumtable.c,v 1.9 2002/08/11 10:39:18 bu Exp $

Copyright 2001-2002, Bulent Yilmaz

Portions Copyright 2002, Michael Heuss

Part of Microbrew MicroSieve
USENET SPAM filter

http://www.microbrew.org/products/usieve.html


MicroSieve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MicroSieve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MicroSieve; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

**********************************************************************/

#include "sumtable.h"
#define TABLEBALANCE 0

#if HISTORY_CHECK

struct node {
    unsigned long long sum;
    int hits;
    struct node *left, *right, *parent;
};

static long tablesize=0;
struct node *nodetable=NULL;
static long nodestep=-1;
static int looped=0;

int tablecheck(unsigned long long checksum) {

    long i;
    static struct node *hashtable[65536];
    struct node *tracer, *lasttrace, *freenode;
    unsigned int hashnode;
    int donetraverse=0;

    /* See if I know how long my sum list is. */
    
    if(tablesize==0) {
#if TESTOUTPUT
        cflog("Getting size of sum list.");
#endif
        if((tablesize=gethistorysize()) == 0) {
#if TESTOUTPUT
            cflog("Could not retrieve history size.  Returning 'Passed'.");
#endif        
            return 1;
        }
    }

    /* See if I've allocated and initialized my nodes table. */
    
    if(nodetable==NULL) {
#if TESTOUTPUT
        cflog("Allocating node table for BST.");
#endif
        nodetable=malloc(tablesize*sizeof(struct node));
        
        if(nodetable==NULL) {
#if TESTOUTPUT
            cflog("Could not allocate memory for node table for BST.  Returning 'Passed'.");
#endif
            return 1;
        }
        
        /* Initialize my hash table. */
        for(i=0; i<65536; i++) hashtable[i]=NULL;
    }
    
    nodestep+=1;
    
    /* See if my table just looped. */
    
    if(nodestep>=tablesize) {
#if TESTOUTPUT
        cflog("Table usage looped.");
#endif
        nodestep=0;
        looped=1;
    }
    
    freenode=&nodetable[nodestep];

    /* Debugging. */
    
    if((looped==1)&&(freenode->hits!=0)) {
        /* Need logic to reclaim nodetable[nodestep]. */

        /* Case 1: Both left and right are unoccupied. */
        
        if((freenode->left==NULL) && (freenode->right==NULL)) {
            /* This is a far leaf.  Just lop it off. */
            
            if(freenode->parent == NULL) {
                /* This is a 'root node'.  Fix the hash table. */
                
                hashnode = (freenode->sum & 0xFFFF000000000000) >> 48;
                
                hashtable[hashnode]=NULL;
                
            } else {
            
                if(freenode->parent->left == freenode) {
                    /* This is to the left of the parent.  Fix the parent. */
                
                    freenode->parent->left=NULL;
                } else if(freenode->parent->right == freenode) {
                    /* This is to the right of the parent.  Fix the parent. */
                
                    freenode->parent->right=NULL;
                } else {
                    /* Just in case. */

                    cflog("Unattached node.  Something is terribly wrong. (Case 1)");
                }
            }
        } /* Case 1. */
        
        /* Case 2: Left node but no right node. */
        
        else if((freenode->left!=NULL) && (freenode->right==NULL)) {
            /* Just remove the node and link it back up to the parent. */
            
            if(freenode->parent == NULL) {
                /* Fix the 'root node' to point at the 'left' node. */
                
                hashnode = (freenode->sum & 0xFFFF000000000000) >> 48;
            
                hashtable[hashnode]=freenode->left;
                
                freenode->left->parent=NULL;
            } else {
                if(freenode->parent->left == freenode) {
                    /* This is to the left of the parent.  Fix the parent. */
                
                    freenode->parent->left=freenode->left;
                    
                    freenode->left->parent=freenode->parent;
                } else if(freenode->parent->right == freenode) {
                    /* This is to the right of the parent.  Fix the parent. */
                
                    freenode->parent->right=freenode->left;
                    
                    freenode->left->parent=freenode->parent;
                } else {
                    /* Just in case. */

                    cflog("Unattached node.  Something is terribly wrong. (Case 2)");
                }
            }
        
        } /* Case 2. */
        
        /* Case 3: Right node but no left node. */
        
        else if((freenode->left==NULL) && (freenode->right!=NULL)) {
            /* Just remove the node and link it back up to the parent. */

            if(freenode->parent==NULL) {
                /* Fix the 'root node' to point at the 'right' node. */
                
                hashnode = (freenode->sum & 0xFFFF000000000000) >> 48;
            
                hashtable[hashnode]=freenode->right;
                
                freenode->right->parent=NULL;
            } else {
                if(freenode->parent->left == freenode) {
                    /* This is to the left of the parent.  Fix the parent. */
                
                    freenode->parent->left=freenode->right;
                    
                    freenode->right->parent=freenode->parent;
                } else if(freenode->parent->right == freenode) {
                    /* This is to the right of the parent.  Fix the parent. */
                
                    freenode->parent->right=freenode->right;
                    
                    freenode->right->parent=freenode->parent;
                } else {
                    /* Just in case. */

                    cflog("Unattached node.  Something is terribly wrong. (Case 3)");
                }
            }
        } /* Case 3. */
        
        /* Case 4: Both right and left nodes are occupied. */
        
        else if((freenode->left!=NULL) && (freenode->right!=NULL)) {
            /* For better balance, first determine if I want (thanks Mike)
                  1. The left's rightmost child, or
                  2. The right's leftmost child.  */
#if TABLEBALANCE                  
            if( (freenode->left->sum - freenode->sum) < (freenode->sum - freenode->right->sum) ) {
#endif            
                /* I am closer to the left's sum.  Favor the right. */

                tracer=freenode->right;
            
                while(tracer->left != NULL) tracer=tracer->left;
            
                /* Now, 'tracer' should be on the right's leftmost child. */
            
                /* Fix up tracer's location. */

                if(tracer!=freenode->right) {
                    tracer->parent->left=tracer->right;
                    if(tracer->right != NULL) tracer->right->parent=tracer->parent;
                }
                
                /* Move tracer to freenode's place. */
                
                if(freenode->parent == NULL) {
                    /* I am at the 'root node'. Fix up the hash table too. */
                    
                    hashnode = (freenode->sum & 0xFFFF000000000000) >> 48;
                    
                    hashtable[hashnode]=tracer;
                    
                    tracer->parent=NULL;
                    tracer->left=freenode->left;
                    tracer->left->parent=tracer;
                    if(tracer!=freenode->right) {
                        tracer->right=freenode->right;
                        freenode->right->parent=tracer;
                    }
                    
                } else {
                    if(freenode->parent->left==freenode) {
                        /* I am to the left of the parent. */
                        
                        freenode->parent->left=tracer;
                        
                        tracer->parent=freenode->parent;
                        tracer->left=freenode->left;
                        tracer->left->parent=tracer;
                        if(tracer!=freenode->right) {
                            tracer->right=freenode->right;
                            freenode->right->parent=tracer;
                        }
                    
                    } else if(freenode->parent->right==freenode) {
                        /* I am to the right of the parent. */
                    
                        freenode->parent->right=tracer;
                        
                        tracer->parent=freenode->parent;
                        tracer->left=freenode->left;
                        tracer->left->parent=tracer;
                        if(tracer!=freenode->right) {
                            tracer->right=freenode->right;
                            freenode->right->parent=tracer;
                        }
                    
                    } else {
                        /* Just in case. */
                        
                        cflog("Unattached node.  Something is terribly wrong. (Case 4a)");
                    }
                }
#if TABLEBALANCE
            } else {
                /* I am closer to the left's sum.  Favor the right. */

                tracer=freenode->left;
            
                while(tracer->right != NULL) tracer=tracer->right;
            
                /* Now, 'tracer' should be on the left's rightmost child. */
            
                /* Fix up tracer's location. */

                if(tracer!=freenode->left) {
                    tracer->parent->right=tracer->left;
                    if(tracer->left != NULL) tracer->left->parent=tracer->parent;
                }
                
                /* Move tracer to freenode's place. */
                
                if(freenode->parent == NULL) {
                    /* I am at the 'root node'. Fix up the hash table too. */
                    
                    hashnode = (freenode->sum & 0xFFFF000000000000) >> 48;
                    
                    hashtable[hashnode]=tracer;
                    
                    tracer->parent=NULL;
                    tracer->right=freenode->right;
                    tracer->right->parent=tracer;
                    if(tracer!=freenode->left) {
                        tracer->left=freenode->left;
                        freenode->left->parent=tracer;
                    }
                    
                } else {
                    if(freenode->parent->left==freenode) {
                        /* I am to the left of the parent. */
                        
                        freenode->parent->left=tracer;
                        
                        tracer->parent=freenode->parent;
                        tracer->right=freenode->right;
                        tracer->right->parent=tracer;
                        if(tracer!=freenode->left) {
                            tracer->left=freenode->left;
                            freenode->left->parent=tracer;
                        }
                    
                    } else if(freenode->parent->right==freenode) {
                        /* I am to the right of the parent. */
                    
                        freenode->parent->right=tracer;
                        
                        tracer->parent=freenode->parent;
                        tracer->right=freenode->right;
                        tracer->right->parent=tracer;
                        if(tracer!=freenode->left) {
                            tracer->left=freenode->left;
                            freenode->left->parent=tracer;
                        }
                    
                    } else {
                        /* Just in case. */
                        
                        cflog("Unattached node.  Something is terribly wrong. (Case 4a)");
                    }
                }
            }
#endif            
        } /* Case 4. */
        
        else cflog("Node doesn't match any of the four cases.");
        
    } else if((looped==1)&&(freenode->hits==0)) cflog("No reclaim needed.");
    
    /* Figure out which 'root node' to use. */
    
    hashnode=(checksum & 0xFFFF000000000000) >> 48;
    
    /* Now find or insert my checksum. */
    
    tracer=hashtable[hashnode];
    
    lasttrace=NULL;
    
    if(tracer==NULL) {
        /* Means I don't have anything for this entry. */
#if TESTOUTPUT
        temphandle=getloghandle();
        fprintf(temphandle, "First entry for hash entry %ud.\n", hashnode);
        fflush(temphandle);
#endif
        freenode->sum=checksum;
        freenode->hits=1;
        freenode->parent=NULL;
        freenode->left=NULL;
        freenode->right=NULL;
        
        hashtable[hashnode]=freenode;
    
    } else {
        /* Time to traverse to find my entry or insert it. */
        
        donetraverse=0;
        
        while(!donetraverse) {
            if(tracer->sum == checksum) {
                /* A hit.  i.e. spam.  Replace this node. */
                
                freenode->sum = checksum;
                freenode->hits = tracer->hits + 1;
                freenode->parent = tracer->parent;
                freenode->left = tracer->left;
                freenode->right = tracer->right;
                
                /* Fix up the parent. */
                
                if(tracer->parent == NULL) {
                    /* I am replacing this 'root node'. */
                
                    hashtable[hashnode]=freenode;
                } else {
                    if(tracer->parent->left==tracer) {
                        /* I am the parent's left node. */
                        
                        tracer->parent->left=freenode;
                        
                    } else {
                        /* I have to be the parent's right. */
                        
                        tracer->parent->right=freenode;   
                    }
                }
                
                /* Fix up the children. */
                
                if(tracer->left != NULL ) tracer->left->parent=freenode;
                if(tracer->right != NULL ) tracer->right->parent=freenode;
                
                tracer->hits=0;
                
                /* Done. */
                
                donetraverse=1;
                
                return 0;
            } else {
            
                if(checksum < tracer->sum) {
                    /* Check to the left. */
                    
                    if(tracer->left != NULL) {
                        lasttrace=tracer;
                        tracer=tracer->left;
                    } else {
                        /* Nothing to the left.  Insert me here. */
                        
                        freenode->parent = tracer;
                        freenode->sum = checksum;
                        freenode->hits = 1;
                        freenode->left = NULL;
                        freenode->right = NULL;
                        
                        tracer->left = freenode;
                        
                        donetraverse=1;
                        
                        return 1;
                    }
                } else if(checksum > tracer->sum) {
                    /* Since checksum is not equal or less, it's greater. */
                    
                    if(tracer->right != NULL) tracer=tracer->right;
                    else {
                        /* Nothing to the right.  Insert me here. */
                        
                        freenode->parent = tracer;
                        freenode->sum = checksum;
                        freenode->hits = 1;
                        freenode->left = NULL;
                        freenode->right = NULL;
                        
                        tracer->right = freenode;
                        
                        donetraverse=1;
                    
                        return 1;
                    }   
                } else {
                    cflog("I couldn't find a place to insert myself.");
                    
                    return 0;
                }
            }
        }
    }
    
    return 1;
    
}

int dumptable() {

    int writedes;
    char *dumpfile=NULL;
    long i;
    
    if((dumpfile=gethistorylogname())!=NULL) {
        writedes=open(dumpfile, O_WRONLY|O_TRUNC|O_CREAT, 0644);
    } else {
        cflog("User did not specify a history log file.");
        return 0;
    }
    
    if(writedes!=-1) {
        if(nodetable!=NULL) {
            /* If it looped, first do the part after 'nodestep' */
        
            if(looped) {
                for(i=nodestep+1; i<tablesize; i++) { 
                    write(writedes, &nodetable[i].sum, sizeof(unsigned long long));
                }
            }
            
            /* Now do the part before nodestep regardless. */
        
            for(i=0; i<=nodestep; i++) {
                if(nodetable[i].hits!=0) write(writedes, &nodetable[i].sum, sizeof(unsigned long long));
            }
        } else cflog("Sum table does not exist.");
    } else return 1;
    
    close(writedes);
    
    return 0;
}

int loadtable() {

    int loaddes;
    int numloads=0;
    char *dumpfile=NULL;
    unsigned long long readdata;

    if((dumpfile=gethistorylogname())!=NULL) {
        loaddes=open(dumpfile, O_RDONLY);
    } else {
        cflog("User did not specify a history log file.");
        return 0;
    }    
    
    if(loaddes!=-1) {
        while(read(loaddes, &readdata, sizeof(unsigned long long))) {
            tablecheck(readdata);
            numloads++;
        }
    } else {
        cflog("Could not open history log for reading.");
        
        return 1;
    }
    
    close(loaddes);

    fprintf(getloghandle(), "Loaded %d checksums from history.\n", numloads);

    return 0;
}

long long calc_sum64(char *instring, long csize) {

    long long sum=0;
    long long *nextptr=NULL;
    long long nextnum;
    double maxsize;
    
    unsigned long int i;
    
    nextptr=(long long *)instring;
    
    if(nextptr) {
        /* Sum only the first 64Kb of the article.  Contributed by mheuss */

        if ((ceil(csize/8)) > (8000.0)) maxsize=8000.0;
        else maxsize=ceil(csize/8);

        /* This is where -funroll-loops comes in handy. */

        for(i=0; i < (maxsize); i++) {
        
            nextnum = (*(nextptr+i));
        
            sum=sum +  ( (nextnum << (sum & 63)) | (nextnum >> ( 64 - (sum & 63) )) );
        }
    }
#if TESTOUTPUT
    else cflog("Pointer did not cast properly.");
#endif
    
    return sum;
}

int summatch(char *instr, long slen) {
    
#if TESTOUTPUT
    FILE *loghandle;
#endif
    long long sum;
    char *alignstr;

#if __SOLARIS2__
    alignstr=memalign( 8, ((ceil(slen/8)+1)*8) );
#else
    alignstr=malloc( ((ceil(slen/8)+1)*8) );
#endif

    memcpy(alignstr, instr, slen);

#if TESTOUTPUT
    cflog("summatch: About to determine article body size");
#endif
    
#if TESTOUTPUT
    loghandle=getloghandle();
    fprintf(loghandle, "summatch: Processing article body.  Size: %ld.\n", slen);
    fflush(loghandle);
#endif
    
    sum=calc_sum64(alignstr, slen);

#if TESTOUTPUT
    loghandle=getloghandle();
    fprintf(loghandle, "summatch: Sum value: %llx.\n", sum);
    fflush(loghandle);
#endif

    free(alignstr);
    
    if(sum) {
        return tablecheck(sum);
    } else return 1;
    
}
#endif /* HISTORY_CHECK */
